# -----------------------------------------------------------------------------
# Copyright (c) 2025- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Global registry of Spyder plugins and decorators for setup and teardown.
"""
