# -----------------------------------------------------------------------------
# Copyright (c) 2020- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
The API-related exception classes emitted by Spyder and its plugins.
"""


class SpyderAPIError(Exception):
    """An error due to incorrect usage by the consumer of the Spyder API."""

    pass
