# -----------------------------------------------------------------------------
# Copyright (c) 2020- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
The gettext-based translation and internationalization (i18n) API for Spyder.
"""

from spyder.config.base import get_translation, _


__all__ = ["get_translation", "_"]
