# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Profiler plugin API.
"""

# Local imports
from spyder.plugins.profiler.widgets.main_widget import (  # noqa
    ProfilerWidgetActions,
    ProfilerWidgetMenus,
    ProfilerContextMenuSections,
    ProfilerWidgetContextMenuActions,
    ProfilerWidgetMainToolbarSections,
)
