# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Debugger Plugin API.
"""

from spyder.plugins.debugger.widgets.main_widget import (
    DebuggerWidgetActions, DebuggerBreakpointActions,
    DebuggerWidgetOptionsMenuSections, DebuggerWidgetMainToolBarSections,
    DebuggerWidgetMenus, DebuggerContextMenuSections
)
