# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
spyder.plugins.remoteclient.api.modules
=======================================

Remote Client Plugin Modules API.

This package contains the API for the Remote Client Plugin
to interact with the Jupyter Server.

These modules provides an interface to control the Jupyter Server that sits
in the remote machine.
"""

from spyder.plugins.remoteclient.api.modules.base import JupyterAPI
from spyder.plugins.remoteclient.api.modules.file_services import (
    SpyderRemoteFileServicesAPI,
)

__all__ = ['JupyterAPI', 'SpyderRemoteFileServicesAPI']
