# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Preferences Plugin API.
"""


class PreferencesActions:
    Show = 'show_action'
    Reset = 'reset_action'
